indexing

   description: "POSIX child processes created by call to exec."

   author: "Berend de Boer"
   date: "$Date: Fri, 13 Oct 2000 13:36:48 +0200 $"
   revision: "$Revision: 1.6 $"

class 

   POSIX_EXEC_PROCESS

inherit
   
   POSIX_FORK_ROOT
      redefine
         stdin,
         stdout,
         stderr,
         fd_stdin,
         fd_stdout,
         fd_stderr
      end

   
creation
   
   make,
   make_capture_input,
   make_capture_output,
   make_capture_io
   
   
feature -- creation
   
   make (a_program: STRING; a_arguments: ARRAY[STRING]) is
      do
         create program_name.make (a_program)
         set_arguments (a_arguments)
      end

   make_capture_input (a_program: STRING; a_arguments: ARRAY[STRING]) is
      do
         make (a_program, a_arguments)
         set_capture_input (True)
      end
   
   make_capture_output (a_program: STRING; a_arguments: ARRAY[STRING]) is
      do
         make (a_program, a_arguments)
         set_capture_output (True)
      end
   
   make_capture_io (a_program: STRING; a_arguments: ARRAY[STRING]) is
      do
         make (a_program, a_arguments)
         set_capture_output (True)
         set_capture_input (True)
      end


feature -- (re)set arguments

   set_arguments (a_arguments: ARRAY[STRING]) is
      do
         if a_arguments = Void then
            create arguments.make(0, -1)
         else
            arguments := a_arguments
         end
      end


feature -- i/o capturing
   
   capture_input: BOOLEAN
         -- is input captured on execute?

   capture_output: BOOLEAN
         -- is output captured on execute?

   capture_error: BOOLEAN
         -- is error captured on execute?

   set_capture_input (on: BOOLEAN) is
      do
         capture_input := on
      end

   set_capture_output (on: BOOLEAN) is
      do
         capture_output := on
      end

   set_capture_error (on: BOOLEAN) is
      do
         capture_error := on
      end


   stdin: POSIX_TEXT_FILE 

   stdout: POSIX_TEXT_FILE
         
   stderr: POSIX_TEXT_FILE
   
   fd_stdin: POSIX_FILE_DESCRIPTOR 

   fd_stdout: POSIX_FILE_DESCRIPTOR
         
   fd_stderr: POSIX_FILE_DESCRIPTOR
   
   
feature -- execute   
   
   exec, execute is
         -- Executes `program_name'
         -- don't forget to `wait' for this process to terminate
      local
         input_pipe,
         output_pipe,
         error_pipe: POSIX_PIPE
         in_child: BOOLEAN
      do
         terminated := False
         
         if capture_input then
            create input_pipe.make
         end
         if capture_output then
            create output_pipe.make
         end
         if capture_error then
            create error_pipe.make
         end
         
         my_pid := posix_fork
         if my_pid = -1 then
            raise_posix_error
         else
            in_child := my_pid = 0
            if in_child  then

               -- this child process has now current process characteristics
               -- open the standard io files if needed
               if capture_input then
                  create fd_stdin.attach_to_fd (STDIN_FILENO)
               end
               if capture_output then
                  create fd_stdout.attach_to_fd (STDOUT_FILENO)
               end
               if capture_error then
                  create fd_stderr.attach_to_fd (STDERR_FILENO)
               end
               
               -- close a captured standard file and attach it to the pipe
               if capture_input then
                  fd_stdin.make_as_duplicate (input_pipe.read_descriptor)
                  check
                     fd_stdin.value = STDIN_FILENO
                  end
               end
               if capture_output then
                  fd_stdout.make_as_duplicate (output_pipe.write_descriptor)
                  check
                     fd_stdout.value = STDOUT_FILENO
                  end
               end
               if capture_error then
                  fd_stderr.make_as_duplicate (error_pipe.write_descriptor)
                  check
                     fd_stderr.value = STDERR_FILENO
                  end
               end

               -- close left over file descriptors
               if capture_input then
                  input_pipe.close
               end
               if capture_output then
                  output_pipe.close
               end
               if capture_error then
                  error_pipe.close
               end
                  
               do_execute

            else -- this is parent
               
               -- make the captured i/o available as the stdout
               -- of this object
               if capture_input then
                  fd_stdin := input_pipe.write_descriptor
                  create stdin.make_from_file_descriptor (fd_stdin, "w")
                  input_pipe.read_descriptor.close
               end
               if capture_output then
                  fd_stdout := output_pipe.read_descriptor
                  create stdout.make_from_file_descriptor (fd_stdout, "r")
                  output_pipe.write_descriptor.close
               end
               if capture_error then
                  fd_stderr := error_pipe.read_descriptor
                  create stderr.make_from_file_descriptor (fd_stderr, "r")
                  error_pipe.write_descriptor.close
               end

            end
         end
      end
   

feature -- accessible state
   
   program_name: POSIX_PATH
         -- program to execute

   arguments: ARRAY[STRING]
         -- arguments to pass to program


feature {NONE}

   do_execute is
      -- actually execute `program_name'
      local
         argv: ARRAY[STRING]
         cargv: ARRAY[POINTER]
         i, j: INTEGER
      do
         create argv.make (0, arguments.count+1)
         argv.put (program_name, 0)
         from
            i := arguments.lower
            j := 1
         until
            i > arguments.upper
         loop
            argv.put (arguments.item(i), j)
            i := i + 1
            j := j + 1
         end
         cargv := ah.string_array_to_pointer_array (argv)
         posix_execvp (sh.string_to_pointer (program_name), 
                       ah.pointer_array_to_pointer (cargv))

         -- we should never return here
         raise_posix_error
      end
   

end -- class POSIX_EXEC_PROCESS
